@@ -1,6 +1,6 @@
 import itertools
 from dataclasses import dataclass
-from typing import Iterable, List, NamedTuple, Optional, Type, TypeVar
+from typing import Iterable, List, NamedTuple, Optional, Type, TypeVar, cast
 
 from loguru import logger
 
@@ -24,17 +24,32 @@ class Edge(NamedTuple):
 class DepNode:
     identifier: int
     form: str
-    lemma: str
-    upos: str
-    xpos: str
-    feats: str
-    head: int
-    deprel: str
-    deps: str
-    misc: str
+    lemma: Optional[str]
+    upos: Optional[str]
+    xpos: Optional[str]
+    feats: Optional[str]
+    head: Optional[int]
+    deprel: Optional[str]
+    deps: Optional[str]
+    misc: Optional[str]
 
     def to_conll(self) -> str:
-        return f"{self.identifier}\t{self.form}\t{self.lemma}\t{self.upos}\t{self.xpos}\t{self.feats}\t{self.head}\t{self.deprel}\t{self.deps}\t{self.misc}"
+        row = [
+            str(c) if c is not None else "_"
+            for c in (
+                self.identifier,
+                self.form,
+                self.lemma,
+                self.upos,
+                self.xpos,
+                self.feats,
+                self.head,
+                self.deprel,
+                self.deps,
+                self.misc,
+            )
+        ]
+        return "\t".join(row)
 
 
 _T_DEPGRAPH = TypeVar("_T_DEPGRAPH", bound="DepGraph")
@@ -54,12 +69,16 @@ class DepGraph:
         self.nodes = list(nodes)
 
         govs = {n.identifier: n.head for n in self.nodes}
-        # FIXME: revise these checks to allow partial supervision
-        if 0 not in govs.values():
-            raise ValueError("Malformed tree: no root")
-        # FIXME: This does not actually ensures that the tree is connex
-        if len(set(govs.values()).difference(govs.keys())) > 1:
-            raise ValueError("Malformed tree: non-connex")
+        # Only do checks on completly annotated trees
+        if None not in govs.values():
+            if 0 not in govs.values():
+                raise ValueError("Malformed tree: no root")
+            if (
+                unreachable_heads := set(govs.values())
+                .difference(govs.keys())
+                .difference((0, None))
+            ):
+                raise ValueError(f"Malformed tree: unreachable heads: {unreachable_heads}")
 
         self.mwe_ranges = [] if mwe_ranges is None else list(mwe_ranges)
         self.metadata = [] if metadata is None else list(metadata)
@@ -72,19 +91,19 @@ class DepGraph:
         return [self.ROOT_TOKEN, *(n.form for n in self.nodes)]
 
     @property
-    def pos_tags(self) -> List[str]:
+    def pos_tags(self) -> List[Optional[str]]:
         """A list where each element list[i] is the upos of the word at position i."""
-        return [self.ROOT_TOKEN, *(n.upos for n in self.nodes)]
+        return [None, *(n.upos for n in self.nodes)]
 
     @property
-    def heads(self) -> List[int]:
+    def heads(self) -> List[Optional[int]]:
         """A list where each element list[i] is the index of the position of the governor of the word at position i."""
-        return [0, *(n.head for n in self.nodes)]
+        return [None, *(n.head for n in self.nodes)]
 
     @property
-    def deprels(self) -> List[str]:
+    def deprels(self) -> List[Optional[str]]:
         """A list where each element list[i] is the dependency label of of the word at position i."""
-        return [self.ROOT_TOKEN, *(n.deprel for n in self.nodes)]
+        return [None, *(n.deprel for n in self.nodes)]
 
     def replace(
         self, edges: Optional[Iterable[Edge]], pos_tags: Optional[Iterable[str]]
@@ -99,9 +118,12 @@ class DepGraph:
         else:
             govs = {e.dep: e.gov for e in edges}
             labels = {e.dep: e.label for e in edges}
+        new_pos_tags: List[Optional[str]]
         if pos_tags is None:
-            pos_tags = self.pos_tags[1:]
-        pos = {i: tag for i, tag in enumerate(pos_tags, start=1)}
+            new_pos_tags = [n.upos for n in self.nodes]
+        else:
+            new_pos_tags = list(pos_tags)
+        pos = {i: tag for i, tag in enumerate(new_pos_tags, start=1)}
         new_nodes = [
             DepNode(
                 identifier=node.identifier,
@@ -136,29 +158,34 @@ class DepGraph:
 
         mwe_ranges = []
         nodes = []
-        for cols in conll:
-            if "-" in cols[0]:
-                mwe_start, mwe_end = cols[0].split("-")
-                mwe_ranges.append(MWERange(int(mwe_start), int(mwe_end), cols[1]))
+        # FIXME: this is clunky, maybe write a better parser that does validation?
+        for row in conll:
+            processed_row: List[Optional[str]]
+            if "-" in row[0]:
+                mwe_start, mwe_end = row[0].split("-")
+                mwe_ranges.append(MWERange(int(mwe_start), int(mwe_end), row[1]))
                 continue
-            if len(cols) < 2:
+            if len(row) < 2:
                 raise ValueError("Too few columns to build a DepNode")
-            elif len(cols) < 10:
-                cols = [*cols, *("_" for _ in range(10 - len(cols)))]
-            if cols[6] == "_":
-                cols[6] = "0"
+            elif len(row) < 10:
+                processed_row = [*row, *("_" for _ in range(10 - len(row)))]
+            else:
+                processed_row = list(row)
+            processed_row[2:] = [c if c != "_" else None for c in processed_row[2:]]
             node = DepNode(
-                identifier=int(cols[0]),
-                form=cols[1],
-                lemma=cols[2],
-                upos=cols[3],
-                xpos=cols[4],
-                feats=cols[5],
-                head=int(cols[6]),
-                deprel=cols[7],
-                deps=cols[8],
-                misc=cols[9],
+                identifier=int(cast(str, processed_row[0])),
+                form=cast(str, processed_row[1]),
+                lemma=processed_row[2],
+                upos=processed_row[3],
+                xpos=processed_row[4],
+                feats=processed_row[5],
+                head=int(processed_row[6]) if processed_row[6] is not None else None,
+                deprel=processed_row[7],
+                deps=processed_row[8],
+                misc=processed_row[9],
             )
+            if node.head is None and node.deprel is not None:
+                logger.warning(f"Node with empty head and nonempty deprel: {node}")
             nodes.append(node)
         return cls(
             nodes=nodes,
@@ -217,14 +244,14 @@ class DepGraph:
                 DepNode(
                     identifier=i,
                     form=w,
-                    lemma="_",
-                    upos="_",
-                    xpos="_",
-                    feats="_",
-                    head=0,
-                    deprel="_",
-                    deps="_",
-                    misc="_",
+                    lemma=None,
+                    upos=None,
+                    xpos=None,
+                    feats=None,
+                    head=None,
+                    deprel=None,
+                    deps=None,
+                    misc=None,
                 )
                 for i, w in enumerate(words, start=1)
             ],
