@@ -13,6 +13,7 @@ from typing import (
     Protocol,
     Sequence,
     Set,
+    Tuple,
     Type,
     TypeVar,
     Union,
@@ -25,7 +26,7 @@ import transformers
 from bidict import bidict, BidirectionalMapping
 from loguru import logger
 from torch import nn
-from torch.nn.utils.rnn import pad_sequence
+from torch.nn.utils.rnn import pack_padded_sequence, pad_sequence
 from transformers.tokenization_utils_base import BatchEncoding, TokenSpan
 
 
@@ -83,6 +84,24 @@ class Lexer(Protocol[_T_LEXER_SENT, _T_LEXER_BATCH]):
         raise NotImplementedError
 
 
+_T_CharRNNLexerBatch = TypeVar("_T_CharRNNLexerBatch", bound="CharRNNLexerBatch")
+
+
+class CharRNNLexerBatch(NamedTuple):
+    encoding: torch.Tensor
+    sent_lengths: List[int]
+    word_lengths: torch.Tensor
+
+    def to(
+        self: _T_CharRNNLexerBatch, device: Union[str, torch.device]
+    ) -> _T_CharRNNLexerBatch:
+        return type(self)(
+            encoding=self.encoding.to(device=device),
+            sent_lengths=self.sent_lengths,
+            word_lengths=self.word_lengths,
+        )
+
+
 _T_CharRNNLexer = TypeVar("_T_CharRNNLexer", bound="CharRNNLexer")
 
 
@@ -130,7 +149,7 @@ class CharRNNLexer(nn.Module):
         self.embedding = nn.Embedding(
             len(self.vocabulary), self.char_embeddings_dim, padding_idx=self.pad_idx
         )
-        # FIXME: this supposes an even output dim
+        # FIXME: this assumes an even output dim
         self.char_bilstm = nn.LSTM(
             batch_first=True,
             bidirectional=True,
@@ -139,21 +158,26 @@ class CharRNNLexer(nn.Module):
             num_layers=1,
         )
 
-    def forward(self, inpt: torch.Tensor) -> torch.Tensor:
+    def forward(self, inpt: CharRNNLexerBatch) -> torch.Tensor:
         """
         Predicts the word embedding from the token characters.
         :param inpt: is a tensor of char indexes encoding a batch of tokens *×token_len
         :return: a word embedding tensor
         """
-        # FIXME: there is probably a better way to do this since this results in tokens that are
-        # full padding We need them of course (they will be cat to other padding embeddings in
-        # graph_parser), but running the RNN on them is frustrating
-        flattened_inputs = inpt.view(-1, inpt.shape[-1])
-        embeddings = self.embedding(flattened_inputs)
-        # ! FIXME: this does not take the padding into account
-        _, (_, cembedding) = self.char_bilstm(embeddings)
-        # TODO: why use the cell state and not the output state here?
-        result = cembedding.view(*inpt.shape[:-1], self.output_dim)
+        embeddings = self.embedding(inpt.encoding)
+        packed_embeddings = pack_padded_sequence(
+            embeddings,
+            lengths=inpt.word_lengths,
+            batch_first=True,
+            enforce_sorted=False,
+        )
+        _, (hn, _) = self.char_bilstm(packed_embeddings)
+        by_sents = (
+            hn.transpose(0, 1)
+            .reshape(inpt.word_lengths.shape[0], self.output_dim)
+            .split(inpt.sent_lengths, dim=0)
+        )
+        result = pad_sequence(by_sents, batch_first=True)
         return result
 
     def word2charcodes(self, token: str) -> torch.Tensor:
@@ -177,37 +201,29 @@ class CharRNNLexer(nn.Module):
                 res = [self.pad_idx]
         return torch.tensor(res, dtype=torch.long)
 
-    def encode(self, tokens_sequence: Sequence[str]) -> torch.Tensor:
+    def encode(self, tokens_sequence: Sequence[str]) -> List[torch.Tensor]:
         """Map word tokens to integer indices.
 
         Returns a tensor of shape `sentence_len×max_word_len`
         """
         subword_indices = [self.word2charcodes(token) for token in tokens_sequence]
-        # shape: sentence_length×num_chars_in_longest_word
-        return pad_sequence(
-            subword_indices, padding_value=self.pad_idx, batch_first=True
-        )
+        return subword_indices
 
-    def make_batch(self, batch: Sequence[torch.Tensor]) -> torch.Tensor:
+    def make_batch(self, batch: Sequence[Sequence[torch.Tensor]]) -> CharRNNLexerBatch:
         """Pad a batch of sentences.
 
-        Returns a tensor of shape `batch_size×max_sentence_len×max_word_len`
+        Returns a `CharRNNLexerBatch` containing flattened, padded encodings (shape:
+        `(batch*max_sent_length, max_word_length)` and a shape to rebuild the per-sentence+per-word
+        encodings.
         """
-        # We need to pad manually because `pad_sequence` only accepts tensors that have all the same
-        # dimensions except for one and we differ both in the sentence lengths dimension and in the
-        # word length dimension
-        res = torch.full(
-            (
-                len(batch),
-                max(t.shape[0] for t in batch),
-                max(t.shape[1] for t in batch),
-            ),
-            fill_value=self.pad_idx,
-            dtype=torch.long,
+        flattened = [encoding for sent in batch for encoding in sent]
+        sent_lengths = [len(sent) for sent in batch]
+        words_lengths = torch.tensor([encoding.shape[0] for encoding in flattened])
+        encoding = pad_sequence(flattened, batch_first=True, padding_value=self.pad_idx)
+
+        return CharRNNLexerBatch(
+            encoding=encoding, sent_lengths=sent_lengths, word_lengths=words_lengths
         )
-        for i, sent in enumerate(batch):
-            res[i, : sent.shape[0], : sent.shape[1]] = sent
-        return res
 
     def save(self, model_path: pathlib.Path, save_weights: bool = True):
         model_path.mkdir(exist_ok=True, parents=True)
@@ -294,18 +310,20 @@ class FastTextLexer(nn.Module):
 
     def forward(self, inpt: torch.Tensor) -> torch.Tensor:
         """
-        :param inpt: a batch of subwords of shape `$*×subwords_dim×features_dim$`
+        :param inpt: a batch of subwords of shape `batch_size×max_sentence_len×max_num_subwords`
         :return: the fasttext embeddings for this batch
         """
         # NOTE: the padding embedding is 0 and should not be modified during training (as per the
-        # `torch.nn.Embedding` doc) so the mean here does not include padding subwords
+        # `torch.nn.Embedding` doc) so the sum here does not include padding subwords
         # NOTE: we use the *mean* of the *input* vectors of the subwords, following [the original
         # FastText
         # implementation](https://github.com/facebookresearch/fastText/blob/a20c0d27cd0ee88a25ea0433b7f03038cd728459/src/fasttext.cc#L117)
         # instead of using the *sum* of *unspecified* (either input or output) vectors as per the
         # original FastText paper (“Enriching Word Vectors with Subword Information”, Bojanowski et
         # al., 2017)
-        return self.embeddings(inpt).mean(dim=-2)
+        # Keepdim here to allow broadcast, clamp to avoid 0/0 for full padding tokens
+        word_lengths = inpt.ne(self.pad_idx).sum(dim=-1, keepdim=True).clamp(min=1)
+        return self.embeddings(inpt).sum(dim=-2) / word_lengths
 
     def word2subcodes(self, token: str) -> torch.Tensor:
         """
@@ -429,25 +447,26 @@ class WordEmbeddingsLexer(nn.Module):
         vocabulary: Sequence[str],
         embeddings_dim: int,
         word_dropout: float,
-        words_padding_idx: int,
         unk_word: str,
     ):
         """Create a new WordEmbeddingsLexer.
 
-        `vocabulary` should not have any duplicates. Note also that the elements at `pad_idx` and
-        `special_tokens idx` will have a special meaning. Building a new `WordEmbeddingsLexer` is
-        preferably done via `from_words`, which takes care of deduplicating and inserting special
+        `vocabulary` should not have any duplicates. Building a new `WordEmbeddingsLexer` is
+        preferably done via `from_words`, which takes care of deduplicating and sorting special
         tokens.
         """
         super().__init__()
+        # Starting at 1 because 0 will be the padding index This is a bit awkward but we can't
+        # really inject a pad token (like `pad` in the char lexer) because any such token might be a
+        # legitimate word so we do this instead.
         try:
             self.vocabulary: BidirectionalMapping[str, int] = bidict(
-                (c, i) for i, c in enumerate(vocabulary)
+                (c, i) for i, c in enumerate(vocabulary, start=1)
             )
         except ValueError as e:
             raise ValueError("Duplicated words in vocabulary") from e
         self.embedding = nn.Embedding(
-            len(self.vocabulary), embeddings_dim, padding_idx=words_padding_idx
+            len(self.vocabulary) + 1, embeddings_dim, padding_idx=0
         )
         self.output_dim = embeddings_dim
         self.unk_word_idx = self.vocabulary[unk_word]
@@ -481,26 +500,23 @@ class WordEmbeddingsLexer(nn.Module):
         return torch.tensor(word_idxes)
 
     def make_batch(self, batch: Sequence[torch.Tensor]) -> torch.Tensor:
-        """Pad a batch of sentences."""
-        return pad_sequence(
-            batch,
-            padding_value=self.embedding.padding_idx,
-            batch_first=True,
-        )
+        """Pad a batch of sentences with zeros."""
+        return pad_sequence(batch, padding_value=0, batch_first=True)
 
     def save(self, model_path: pathlib.Path, save_weights: bool = True):
         model_path.mkdir(exist_ok=True, parents=True)
         config_file = model_path / "config.json"
         with open(config_file, "w") as out_stream:
+            # FIXME: range(1, len())… here is inelegant and is unnecessary if the inverse has ordered values
             json.dump(
                 {
                     "embeddings_dim": self.output_dim,
                     "unk_word": self.vocabulary.inverse[self.unk_word_idx],
                     "vocabulary": [
-                        self.vocabulary.inverse[i] for i in range(len(self.vocabulary))
+                        self.vocabulary.inverse[i]
+                        for i in range(1, len(self.vocabulary) + 1)
                     ],
                     "word_dropout": self.word_dropout,
-                    "words_padding_idx": self.embedding.padding_idx,
                 },
                 out_stream,
             )
@@ -519,6 +535,7 @@ class WordEmbeddingsLexer(nn.Module):
             res.load_state_dict(torch.load(weight_file, map_location="cpu"))
         return res
 
+    # FIXME: probably add thresholds here, at least to filter out hapax?
     @classmethod
     def from_words(
         cls: Type[_T_WordEmbeddingsLexer], words: Iterable[str], **kwargs
@@ -558,7 +575,6 @@ _T_BertLexerBatch = TypeVar("_T_BertLexerBatch", bound="BertLexerBatch")
 
 
 class BertLexerBatch(NamedTuple):
-    word_indices: torch.Tensor
     encoding: BatchEncoding
     subword_alignments: Sequence[Sequence[TokenSpan]]
 
@@ -566,7 +582,6 @@ class BertLexerBatch(NamedTuple):
         self: _T_BertLexerBatch, device: Union[str, torch.device]
     ) -> _T_BertLexerBatch:
         return type(self)(
-            self.word_indices.to(device=device),
             self.encoding.to(device=device),
             self.subword_alignments,
         )
@@ -670,6 +685,7 @@ class BertLexer(nn.Module):
     def forward(self, inpt: BertLexerBatch) -> torch.Tensor:
         layers = self.model(
             input_ids=inpt.encoding["input_ids"],
+            attention_mask=inpt.encoding["attention_mask"],
             return_dict=True,
             output_hidden_states=True,
         ).hidden_states
@@ -697,14 +713,12 @@ class BertLexer(nn.Module):
                 subword_embeddings.shape[2],
             )
         )
-        word_embeddings[:, 0, ...] = subword_embeddings.mean(dim=1)
         # FIXME: this loop is embarassingly parallel, there must be a way to parallelize it
         for sent_n, alignment in enumerate(inpt.subword_alignments):
             # TODO: If we revise the alignment format, this could probably be made faster using
             # <https://pytorch.org/docs/stable/generated/torch.scatter.htmlW or
             # <https://pytorch-scatter.readthedocs.io/en/latest/functions/scatter.html>
-            # The word indices start at 1 because word 0 is the root token, for which we have no
-            # bert embedding so we use the average of all subword embeddings
+            # The word indices start at 1 because word 0 is the root token
             for word_n, span in enumerate(alignment, start=1):
                 # shape: `span.end-span.start×features`
                 this_word_subword_embeddings = subword_embeddings[
@@ -719,6 +733,10 @@ class BertLexer(nn.Module):
                 else:
                     raise ValueError(f"Unknown reduction {self.subwords_reduction}")
                 word_embeddings[sent_n, word_n, ...] = reduced_bert_word_embedding
+            # Now do the root token as the average of (non-padding words) embeddings.
+            word_embeddings[sent_n, 0, ...] = word_embeddings[sent_n, ...].sum(
+                dim=0
+            ) / len(alignment)
 
         return word_embeddings
 
@@ -734,7 +752,6 @@ class BertLexer(nn.Module):
         encoding = self.tokenizer.pad(encodings_lst)
         encoding.convert_to_tensors("pt")
         return BertLexerBatch(
-            torch.zeros((1,)),
             encoding,
             subwords_alignments,
         )
