@@ -6,11 +6,11 @@ import random
 import shutil
 import warnings
 from typing import (
-    Any,
-    Dict,
     IO,
+    Any,
     BinaryIO,
     Callable,
+    Dict,
     Final,
     Iterable,
     List,
@@ -32,22 +32,21 @@ import pydantic
 import torch
 import transformers
 import yaml
-from bidict import bidict, BidirectionalMapping
+from bidict import BidirectionalMapping, bidict
 from boltons import iterutils as itu
 from loguru import logger
 from torch import nn
 from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence, pad_sequence
 
-from hopsparser import deptree, lexers
-from hopsparser import utils
+from hopsparser import deptree, lexers, utils
 from hopsparser.deptree import DepGraph
 from hopsparser.lexers import (
     BertLexer,
     CharRNNLexer,
+    FastTextLexer,
     Lexer,
     SupportsTo,
     WordEmbeddingsLexer,
-    FastTextLexer,
     freeze_module,
 )
 from hopsparser.mst import chuliu_edmonds_one_root as chuliu_edmonds
@@ -55,7 +54,9 @@ from hopsparser.utils import smart_open
 
 
 def gen_tags(treelist: Iterable[DepGraph]) -> List[str]:
-    tagset = set([tag for tree in treelist for tag in tree.pos_tags[1:]])
+    tagset = set(
+        [tag for tree in treelist for tag in tree.pos_tags[1:] if tag is not None]
+    )
     if tagset.intersection((BiAffineParser.PAD_TOKEN, BiAffineParser.UNK_WORD)):
         raise ValueError("Tag conflict: the treebank contains reserved POS tags")
     return [
@@ -67,12 +68,13 @@ def gen_tags(treelist: Iterable[DepGraph]) -> List[str]:
 
 
 def gen_labels(treelist: Iterable[DepGraph]) -> List[str]:
-    labels = set([lbl for tree in treelist for lbl in tree.deprels])
+    labels = set([lbl for tree in treelist for lbl in tree.deprels if lbl is not None])
     if BiAffineParser.PAD_TOKEN in labels:
         raise ValueError("Tag conflict: the treebank contains reserved dep labels")
     return [BiAffineParser.PAD_TOKEN, *sorted(labels)]
 
 
+# FIXME: why does this not have relu in output????
 class MLP(nn.Module):
     def __init__(
         self, input_dim: int, hidden_dim: int, output_dim: int, dropout: float = 0.0
@@ -137,39 +139,41 @@ class SentencesBatch(NamedTuple):
 
     - `words` The word forms for every sentence in the batch
     - `encodings` A dict mapping a lexer name to the correponding encoding of the batch of sentences
+    - `sent_length` The lengths of the sentences in the batch as `LongTensor` with shape
+      `(batch_size,)`
+    """
+
+    words: Sequence[Sequence[str]]
+    encodings: Dict[str, SupportsTo]
+    sent_lengths: torch.Tensor
+
+    def to(
+        self: _T_SentencesBatch, device: Union[str, torch.device]
+    ) -> _T_SentencesBatch:
+        return type(self)(
+            encodings={
+                lexer_name: encoded_batch.to(device)
+                for lexer_name, encoded_batch in self.encodings.items()
+            },
+            sent_lengths=self.sent_lengths,
+            words=self.words,
+        )
+
+
+class EncodedTree(NamedTuple):
+    """Annotations for an `EncodedSentence`.
+
+    ## Attributes
+
+    - `sentence`: the sentence in question
     - `tags` The gold POS tags (if any) as a `LongTensor` with shape `(batch_size,
       max_sentence_length)`
     - `heads` The gold heads (if any) as a `LongTensor` with shape `(batch_size,
       max_sentence_length)`
     - `labels` The gold dependency labels (if any) as a `LongTensor` with shape `(batch_size,
       max_sentence_length)`
-    - `sent_length` The lengths of the sentences in the batch as `LongTensor` with shape
-      `(batch_size,)`
-    - `content_mask` A `BoolTensor` mask of shape `(batch_size, max_sentence_length)` such that
-      `content_mask[i, j]` is true iff the j-th word of the i-th sentence in the batch is neither
-      padding not the root (i.e. iff `1 <= j < sent_length[i]`).
     """
 
-    words: Sequence[Sequence[str]]
-    encodings: Dict[str, SupportsTo]
-    sent_lengths: torch.Tensor
-    content_mask: torch.Tensor
-
-    def to(
-        self: _T_SentencesBatch, device: Union[str, torch.device]
-    ) -> _T_SentencesBatch:
-        return type(self)(
-            words=self.words,
-            encodings={
-                lexer_name: encoded_batch.to(device)
-                for lexer_name, encoded_batch in self.encodings.items()
-            },
-            sent_lengths=self.sent_lengths,
-            content_mask=self.content_mask.to(device),
-        )
-
-
-class EncodedTree(NamedTuple):
     sentence: EncodedSentence
     heads: torch.Tensor
     labels: torch.Tensor
@@ -278,7 +282,7 @@ class BiAffineParser(nn.Module):
         # The issue is `Lexer` is a protocol and we can't require them to be torch Modules (which is
         # a concrete class and you can't ask a protocol to subclass a concrete class or ask for the
         # intersection of a protocol and a concrete class), but we do need them to be torch Modules
-        # since the have to wrap them in a `ModuleDict` here`. Also we really don't want to separate
+        # since we have to wrap them in a `ModuleDict` here`. Also we really don't want to separate
         # lexers that wouldn't be torch Modules since that's really never going to happen in
         # practice.
         self.lexers = cast(
@@ -381,23 +385,22 @@ class BiAffineParser(nn.Module):
         batch_size = batch.heads.shape[0]
         num_padded_deps = batch.heads.shape[1]
         num_deprels = lab_scores.shape[-1]
-        # ARC LOSS
-        # [batch×num_deps, num_heads]
-        arc_scores_flat = arc_scores.view(-1, arc_scores.size(-1))
-        arc_loss = marginal_loss(arc_scores_flat, batch.heads.view(-1))
 
         # TAGGER_LOSS
         # [batch×sent_length, num_tags]
         tagger_scores_flat = tagger_scores.view(-1, tagger_scores.size(-1))
         tagger_loss = marginal_loss(tagger_scores_flat, batch.tags.view(-1))
 
-        # LABEL LOSS
-        # We will select the labels for the true heads, so we have to give a true head to
-        # the padding tokens (even if they will be ignored in the crossentropy since the true
-        # label for that head is set to -100) so we give them the root.
-        positive_heads = batch.heads.masked_fill(
-            batch.sentences.content_mask.logical_not(), 0
-        )
+        # ARC LOSS
+        # [batch×num_deps, num_heads]
+        arc_scores_flat = arc_scores.view(-1, arc_scores.size(-1))
+        arc_loss = marginal_loss(arc_scores_flat, batch.heads.view(-1))
+
+        # LABEL LOSS We will select the labels for the gold heads, so we have to provide one when
+        # there is none (either it is absent from the data or this is a padding token) to even if
+        # they will be ignored in the crossentropy since the true label for that head is set to -100
+        # so we give them the root.
+        positive_heads = batch.heads.masked_fill(batch.heads.eq(self.LABEL_PADDING), 0)
         heads_selection = positive_heads.view(batch_size, num_padded_deps, 1, 1)
         # [batch, n_dependents, 1, n_labels]
         heads_selection = heads_selection.expand(
@@ -430,12 +433,13 @@ class BiAffineParser(nn.Module):
         tag_tp = torch.zeros(1, dtype=torch.long, device=device)
         arc_tp = torch.zeros(1, dtype=torch.long, device=device)
         lab_tp = torch.zeros(1, dtype=torch.long, device=device)
-        overall_size = torch.zeros(1, dtype=torch.long, device=device)
+        overall_tag_size = torch.zeros(1, dtype=torch.long, device=device)
+        overall_arc_size = torch.zeros(1, dtype=torch.long, device=device)
+        overall_lab_size = torch.zeros(1, dtype=torch.long, device=device)
         gloss = torch.zeros(1, dtype=torch.float, device=device)
 
         with torch.inference_mode():
             for batch in dev_set:
-                overall_size += batch.sentences.content_mask.sum()
 
                 batch = batch.to(device)
 
@@ -453,29 +457,23 @@ class BiAffineParser(nn.Module):
                     tagger_scores, arc_scores, lab_scores, batch, loss_fnc
                 )
 
-                # greedy arc accuracy (without parsing)
-                arc_pred = arc_scores.argmax(dim=-1)
-                arc_accuracy = (
-                    arc_pred.eq(batch.heads)
-                    .logical_and(batch.sentences.content_mask)
-                    .sum()
-                )
-                arc_tp += arc_accuracy
-
                 # tagger accuracy
                 tag_pred = tagger_scores.argmax(dim=-1)
-                tag_accuracy = (
-                    tag_pred.eq(batch.tags)
-                    .logical_and(batch.sentences.content_mask)
-                    .sum()
-                )
+                tags_mask = batch.tags.ne(self.LABEL_PADDING)
+                overall_tag_size += tags_mask.sum()
+                tag_accuracy = tag_pred.eq(batch.tags).logical_and(tags_mask).sum()
                 tag_tp += tag_accuracy
 
+                # greedy arc accuracy (without parsing)
+                arc_pred = arc_scores.argmax(dim=-1)
+                arc_mask = batch.heads.ne(self.LABEL_PADDING)
+                overall_arc_size += arc_mask.sum()
+                arc_accuracy = arc_pred.eq(batch.heads).logical_and(arc_mask).sum()
+                arc_tp += arc_accuracy
+
                 # greedy label accuracy (without parsing)
                 gold_heads_select = (
-                    batch.heads.masked_fill(
-                        batch.sentences.content_mask.logical_not(), 0
-                    )
+                    batch.heads.masked_fill(batch.heads.eq(self.LABEL_PADDING), 0)
                     .view(batch.heads.shape[0], batch.heads.shape[1], 1, 1)
                     .expand(
                         batch.heads.shape[0],
@@ -489,18 +487,18 @@ class BiAffineParser(nn.Module):
                     lab_scores, -2, gold_heads_select
                 ).squeeze(-2)
                 lab_pred = gold_head_lab_scores.argmax(dim=-1)
-                lab_accuracy = (
-                    lab_pred.eq(batch.labels)
-                    .logical_and(batch.sentences.content_mask)
-                    .sum()
-                )
+                lab_mask = batch.labels.ne(self.LABEL_PADDING)
+                overall_lab_size += lab_mask.sum()
+                lab_accuracy = lab_pred.eq(batch.labels).logical_and(lab_mask).sum()
                 lab_tp += lab_accuracy.item()
 
         return (
-            gloss.true_divide(overall_size).item(),
-            tag_tp.true_divide(overall_size).item(),
-            arc_tp.true_divide(overall_size).item(),
-            lab_tp.true_divide(overall_size).item(),
+            gloss.true_divide(
+                overall_tag_size + overall_arc_size + overall_lab_size
+            ).item(),
+            tag_tp.true_divide(overall_tag_size).item(),
+            arc_tp.true_divide(overall_arc_size).item(),
+            lab_tp.true_divide(overall_lab_size).item(),
         )
 
     def train_model(
@@ -556,7 +554,11 @@ class BiAffineParser(nn.Module):
             )
             self.train()
             for batch in train_batches:
-                overall_size += batch.sentences.content_mask.sum()
+                overall_size += (
+                    batch.tags.ne(self.LABEL_PADDING).sum()
+                    + batch.heads.ne(self.LABEL_PADDING).sum()
+                    + batch.labels.ne(self.LABEL_PADDING).sum()
+                )
 
                 batch = batch.to(device)
 
@@ -579,29 +581,41 @@ class BiAffineParser(nn.Module):
                 scheduler.step()
 
             if dev_set is not None:
-                dev_loss, dev_tag_acc, dev_arc_acc, dev_lab_acc = self.eval_model(
+                dev_loss, dev_tag_acc, dev_head_acc, dev_lab_acc = self.eval_model(
                     dev_set.make_batches(
                         batch_size, shuffle_batches=False, shuffle_data=False
                     ),
                     batch_size=batch_size,
                 )
+                # FIXME: this is not very elegant
                 log_epoch(
                     str(e),
                     {
                         "train loss": f"{train_loss.true_divide(overall_size).item():.4f}",
                         "dev loss": f"{dev_loss:.4f}",
-                        "dev tag acc": f"{dev_tag_acc:06.2%}",
-                        "dev head acc": f"{dev_arc_acc:06.2%}",
-                        "dev deprel acc": f"{dev_lab_acc:06.2%}",
+                        **{
+                            k: f"{v:06.2%}"
+                            for k, v in (
+                                ("dev tag acc", dev_tag_acc),
+                                ("dev head acc", dev_head_acc),
+                                ("dev deprel acc", dev_lab_acc),
+                            )
+                            if not math.isnan(v)
+                        },
                     },
                 )
 
-                if dev_arc_acc > best_arc_acc:
+                # FIXME: probably change the logic here, esp. for headless data
+                if dev_head_acc > best_arc_acc:
                     logger.info(
-                        f"New best model: head accuracy {dev_arc_acc:.2%} > {best_arc_acc:.2%}"
+                        f"New best model: head accuracy {dev_head_acc:.2%} > {best_arc_acc:.2%}"
                     )
                     self.save_params(weights_file)
-                    best_arc_acc = dev_arc_acc
+                    best_arc_acc = dev_head_acc
+                elif math.isnan(dev_head_acc):
+                    logger.debug("No head annotations in dev: saving model")
+                    self.save_params(weights_file)
+                    best_arc_acc = dev_head_acc
             else:
                 self.save_params(weights_file)
 
@@ -656,36 +670,37 @@ class BiAffineParser(nn.Module):
             [sent.sent_len for sent in sentences], dtype=torch.long
         )
 
-        content_mask = (
-            torch.arange(sent_lengths.max().item())
-            .unsqueeze(0)
-            .lt(sent_lengths.unsqueeze(1))
-        )
-        # For the root tokens
-        content_mask[:, 0] = False
-
         return SentencesBatch(
             words=words,
             encodings=encodings,
-            content_mask=content_mask,
             sent_lengths=sent_lengths,
         )
 
     def encode_tree(self, tree: DepGraph) -> EncodedTree:
         sentence = self.encode_sentence(tree.words[1:])
         tag_idxes = torch.tensor(
-            [self.tagset.get(tag, self.LABEL_PADDING) for tag in tree.pos_tags],
+            [
+                self.tagset.get(tag, self.LABEL_PADDING)
+                if tag is not None
+                else self.LABEL_PADDING
+                for tag in tree.pos_tags
+            ],
+            dtype=torch.long,
+        )
+        heads = torch.tensor(
+            [h if h is not None else self.LABEL_PADDING for h in tree.heads],
             dtype=torch.long,
         )
-        tag_idxes[0] = self.LABEL_PADDING
-        heads = torch.tensor(tree.heads, dtype=torch.long)
-        heads[0] = self.LABEL_PADDING
         # FIXME: should unk labels be padding?
         labels = torch.tensor(
-            [self.labels.get(lab, self.LABEL_PADDING) for lab in tree.deprels],
+            [
+                self.labels.get(lab, self.LABEL_PADDING)
+                if lab is not None and h is not None
+                else self.LABEL_PADDING
+                for h, lab in zip(tree.heads, tree.deprels)
+            ],
             dtype=torch.long,
         )
-        labels[0] = self.LABEL_PADDING
         return EncodedTree(
             sentence=sentence,
             heads=heads,
@@ -776,47 +791,56 @@ class BiAffineParser(nn.Module):
                     tagger_scores = tagger_scores[:sentence_length, :]
                     arc_scores = arc_scores[:sentence_length, :sentence_length]
                     lab_scores = lab_scores[:sentence_length, :sentence_length, :]
-                    # Predict heads
-                    probs = arc_scores.cpu().numpy()
-                    mst_heads_np = (
-                        np.argmax(probs, axis=1) if greedy else chuliu_edmonds(probs)
-                    )
-                    # shape: num_deps
-                    mst_heads = torch.from_numpy(mst_heads_np).to(device)
-
-                    # Predict tags
-                    tag_idxes = tagger_scores.argmax(dim=1)
-                    # Predict labels
-                    # shape: num_deps×1×num_deprel
-                    select = mst_heads.view(-1, 1, 1).expand(-1, 1, lab_scores.size(-1))
-                    # shape: num_deps×num_deprel
-                    selected = torch.gather(lab_scores, 1, select).view(
-                        sentence_length, -1
-                    )
-                    mst_labels = selected.argmax(dim=-1)
-
-                    # `[1:]` to ignore the root node's tag
-                    pos_tags = [
-                        self.tagset.inverse[idx] for idx in tag_idxes[1:].tolist()
-                    ]
-                    # `[1:]` to ignore the root node's head
-                    heads = mst_heads[1:].tolist()
-                    # `[1:]` to ignore the root node's deprel
-                    deprels = [
-                        self.labels.inverse[lbl] for lbl in mst_labels[1:].tolist()
-                    ]
-                    edges = [
-                        deptree.Edge(head_idx, lbl, dep_idx)
-                        for dep_idx, (head_idx, lbl) in enumerate(
-                            zip(heads, deprels), start=1
-                        )
-                    ]
-                    result_tree = tree.replace(
-                        edges=edges,
-                        pos_tags=pos_tags,
+                    result_tree = self._scores_to_tree(
+                        arc_scores=arc_scores,
+                        greedy=greedy,
+                        lab_scores=lab_scores,
+                        tagger_scores=tagger_scores,
+                        tree=tree,
                     )
                     yield result_tree
 
+    def _scores_to_tree(
+        self,
+        arc_scores: torch.Tensor,
+        greedy: bool,
+        lab_scores: torch.Tensor,
+        tagger_scores: torch.Tensor,
+        tree: DepGraph,
+    ):
+        sentence_length = lab_scores.shape[0]
+        # Predict heads
+        probs = arc_scores.cpu().numpy()
+        mst_heads_np = np.argmax(probs, axis=1) if greedy else chuliu_edmonds(probs)
+        # shape: num_deps
+        mst_heads = torch.from_numpy(mst_heads_np).to(lab_scores.device)
+
+        # Predict tags
+        tag_idxes = tagger_scores.argmax(dim=1)
+        # Predict labels
+        # shape: num_deps×1×num_deprel
+        select = mst_heads.view(-1, 1, 1).expand(-1, 1, lab_scores.size(-1))
+        # shape: num_deps×num_deprel
+        selected = torch.gather(lab_scores, 1, select).view(sentence_length, -1)
+        mst_labels = selected.argmax(dim=-1)
+
+        # `[1:]` to ignore the root node's tag
+        pos_tags = [self.tagset.inverse[idx] for idx in tag_idxes[1:].tolist()]
+        # `[1:]` to ignore the root node's head
+        heads = mst_heads[1:].tolist()
+        # `[1:]` to ignore the root node's deprel
+        deprels = [self.labels.inverse[lbl] for lbl in mst_labels[1:].tolist()]
+        edges = [
+            deptree.Edge(head_idx, lbl, dep_idx)
+            for dep_idx, (head_idx, lbl) in enumerate(zip(heads, deprels), start=1)
+        ]
+        result_tree = tree.replace(
+            edges=edges,
+            pos_tags=pos_tags,
+        )
+
+        return result_tree
+
     # FIXME: this is awkward when parsing pre-tokenized input: we should accept something like `inpt:
     # Iterable[Sequence[str]]`
     def parse(
